#!/usr/bin/env python
# Written by Alan Milligan and Danica Sutherland (2023-24)
# Based on CPSC 540 Julia code by Mark Schmidt
# and some CPSC 340 Python code by Mike Gelbart and Nam Hee Kim, among others

import os
from pathlib import Path
import pickle

import numpy as np
from scipy.special import betaln
from sklearn.neighbors import KNeighborsClassifier


# make sure we're working in the directory this file lives in,
# for simplicity with imports and relative paths
os.chdir(Path(__file__).parent.resolve())

# question code
from generative_classifiers import GDA, TDA
from utils import (
    test_and_plot,
    load_dataset,
    main,
    handle,
    run,
)


def bernoulli_nll(theta, n_pos, n_total):
    n_neg = n_total - n_pos
    return -n_pos * np.log(theta) - n_neg * np.log1p(-theta)
    # np.log1p(x) == np.log(1 + x)  except that log1p is more accurate for small |x|


def berns_nll(thetas, data):
    return sum(
        bernoulli_nll(theta, n_pos, n_total)
        for theta, (n_pos, n_total) in zip(thetas, data)
    )


@handle("eb-base")
def eb_base():
    n, n_test = load_dataset("cancer_data", "n", "n_test")
    n_groups = n.shape[0]

    theta = np.full(n_groups, 0.5)
    print(f"NLL for theta = 0.5: {berns_nll(theta, n_test) : .1f}")

    mle_thetas = n[:, 0] / n[:, 1]
    with np.errstate(all="ignore"):  # ignore numerical warnings about nans
        print(f"NLL for theta = MLE: {berns_nll(mle_thetas, n_test) : .1f}")


@handle("eb-map")
def eb_map():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


@handle("eb-bayes")
def eb_bayes():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


@handle("eb-pooled")
def eb_pooled():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


@handle("eb-max")
def eb_max():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


@handle("eb-newprior")
def eb_newprior():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


@handle("eb-newprior-sep")
def eb_newprior_sep():
    n, n_test = load_dataset("cancer_data", "n", "n_test")

    raise NotImplementedError()


################################################################################


def eval_models(models, ds_name):
    X, y, X_test, y_test = load_dataset(ds_name, "X", "y", "X_test", "y_test")
    for model in models:
        model.fit(X, y)
        y_hat = model.predict(X_test)
        yield np.mean(y_hat != y_test)


def eval_model(model, ds_name):
    return next(eval_models([model], ds_name))


@handle("gda")
def gda():
    model = KNeighborsClassifier(n_neighbors=1)
    print(f"{model.n_neighbors}-NN test error: {eval_model(model, 'gauss_noise'):.1%}")

    model = GDA()
    print(f"GDA  test error: {eval_model(model, 'gauss_noise'):.1%}")


@handle("tda")
def tda():
    model = TDA()
    print(f"TDA  test error: {eval_model(model, 'gauss_noise'):.1%}")


if __name__ == "__main__":
    main()
